/****************************************************************************/
/*                                                                          */
/* Program name: MQIsdp protocol sample application                         */
/*                                                                          */
/* Description: If the demo application is running then this application    */
/*              can be used to update the bounds within which the demo      */
/*              application operates.                                       */
/*                                                                          */
/*  Statement:  Licensed Materials - Property of IBM                        */
/*                                                                          */
/*              MQSeries SupportPac IA93                                    */
/*              (C) Copyright IBM Corp. 2002                                */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/* Function:                                                                */
/*                                                                          */
/* This application simply publishes a parameter string to topic            */
/* demo/c/control. If the demo application is running it receives the       */
/* publication and dynamically updates the bounds it is running in.         */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/* Change history:                                                          */
/*                                                                          */
/* V1.0   19-02-2003  IRH  Initial release                                  */
/*                                                                          */
/*==========================================================================*/
/* Module Name: democ.c                                                      */
#include <stdio.h>
#ifdef WIN32
  #include <time.h>
  #include <windows.h>
#else  
  #include <sys/time.h>
  #include <unistd.h>
  #include <stdarg.h>
  #include <sys/types.h>
  #include <sys/ipc.h>
  #include <sys/sem.h>
  #include <pthread.h>
  
#endif
#include <MQIsdp.h>

#define CaseRet( x ) \
  case x:            \
    ptr = #x;        \
    break;

typedef struct struc_mqisdp {
    MQISDPCH  hConn;
    char      infoString[MQISDP_INFO_STRING_LENGTH];
    long      infoCode;
    MQISDPTI  hMbs;
} MQISDP_PARMS;

typedef struct struc_ppp {
    char  *ControlTopic;
    char   IpAddress[MQISDP_INET_ADDR_LENGTH];
    long   IpPort;
    MQISDP_PARMS MqisdpParms;
    long   NumPairs;
    char   Pairs[64];
} PROG_PARMS;
    
#define CLIENT_ID "MQISDP_C_DEMOCTRL"
static int demoUsage( void );
static void demoLog( char *fmt, ... );
static void demoSleep( int mSecs );
static int demoInitialise( PROG_PARMS *ppp );
static int demoProcessCommandLine( PROG_PARMS *ppp, int argCount, char *argList[] );
static int demoMQIsdpConnect( PROG_PARMS *ppp );
static int demoMQIsdpDisconnect( PROG_PARMS *ppp );
static int demoMQIsdpPublish( PROG_PARMS *ppp );
static int demoGetRCtoString( long rc, char *pBuffer, long bufSz );

int main( int argc, char **argv ) {
    MQISDPTI  *pSendTaskParms;
    MQISDPTI  *pRcvTaskParms;
    PROG_PARMS pp;

    demoInitialise( &pp );

    if ( demoProcessCommandLine( &pp, argc, argv ) != 0 ) {
        exit(1);
    }
    
    /* Send and receive task parms must be on the heap */
    pSendTaskParms = (MQISDPTI*)malloc( sizeof(MQISDPTI) );
    pRcvTaskParms = (MQISDPTI*)malloc( sizeof(MQISDPTI) );
    if ( MQIsdp_StartTasks( &pp.MqisdpParms.hMbs, pSendTaskParms,
                            pRcvTaskParms, CLIENT_ID ) != 0 ) {
        exit(1);
    }
    
    if ( pp.NumPairs > 0 ) {
        /* Connect to the MQIsdp broker */
        if ( demoMQIsdpConnect( &pp ) != 0 ) {
            exit(2);
        }

        demoMQIsdpPublish( &pp );

        demoMQIsdpDisconnect( &pp );
    }

    return 0;
}

/* All intervals are in seconds */
static int demoInitialise( PROG_PARMS *ppp ) {

    ppp->ControlTopic = "demo/c/control";
    ppp->IpPort = 1883;
    ppp->NumPairs = 0;
    ppp->Pairs[0] = '\0';
    ppp->MqisdpParms.hConn = MQISDP_INV_CONN_HANDLE;
    return 0;
}

static int demoProcessCommandLine( PROG_PARMS *ppp, int argCount, char *argList[] ) {
    int   i;
    int   rc = 2;
    int   first = 1;
    char *insert = " ";
    /* Set to 0 if mandatory parm -a is specified */
    /* Set to 1 on error                          */

    for ( i=1; i < argCount; i++ ) {
        switch ( argList[i][1] ) {
        case 'm':
            /* Mean flow rate (seconds) */
            if ( strcmp( argList[i], "-m" ) == 0 ) {
                i++;
                if ( i < argCount ) {
                    ppp->NumPairs++;
                    strcat( ppp->Pairs," m " );
                    strcat( ppp->Pairs, argList[i] );
                } else {
                    rc = 1;
                    demoUsage();
                }
            } else {
                demoUsage();
                rc = 1;
            }
            break;
        case 'x':
            /* Max flow rate (seconds) */
            if ( strcmp( argList[i], "-x" ) == 0 ) {
                i++;
                if ( i < argCount ) {
                    ppp->NumPairs++;
                    strcat( ppp->Pairs," x " );
                    strcat( ppp->Pairs, argList[i] );
                } else {
                    rc = 1;
                    demoUsage();
                }
            } else {
                demoUsage();
                rc = 1;
            }
            break;
        case 'l':
            /* Min flow rate (seconds) */
            if ( strcmp( argList[i], "-l" ) == 0 ) {
                i++;
                if ( i < argCount ) {
                    ppp->NumPairs++;
                    strcat( ppp->Pairs," l " );
                    strcat( ppp->Pairs, argList[i] );
                } else {
                    rc = 1;
                    demoUsage();
                }
            } else {
                demoUsage();
                rc = 1;
            }
            break;
        case 'r':
            /* Rate change interval (seconds) */
            if ( strcmp( argList[i], "-r" ) == 0 ) {
                i++;
                if ( i < argCount ) {
                    ppp->NumPairs++;
                    strcat( ppp->Pairs," r " );
                    strcat( ppp->Pairs, argList[i] );
                } else {
                    rc = 1;
                    demoUsage();
                }
            } else {
                demoUsage();
                rc = 1;
            }
            break;
        case 'p':
            /* Publish interval (seconds) */
            if ( strcmp( argList[i], "-p" ) == 0 ) {
                i++;
                if ( i < argCount ) {
                    ppp->NumPairs++;
                    strcat( ppp->Pairs," p " );
                    strcat( ppp->Pairs, argList[i] );
                } else {
                    rc = 1;
                    demoUsage();
                }
            } else {
                demoUsage();
                rc = 1;
            }
            break;
        case 'a':
            /* IP address */
            if ( strcmp( argList[i], "-a" ) == 0 ) {
                rc = 0;
                i++;
                if ( i < argCount ) {
                    strncpy( ppp->IpAddress, argList[i], MQISDP_INET_ADDR_LENGTH );
                } else {
                    rc = 1;
                    demoUsage();
                }
            } else {
                demoUsage();
                rc = 1;
            }
            break;
        case 's':
            /* Ip port number */
            if ( strcmp( argList[i], "-s" ) == 0 ) {
                i++;
                if ( i < argCount ) {
                    ppp->IpPort = atol( argList[i] );
                } else {
                    rc = 1;
                    demoUsage();
                }
            } else {
                demoUsage();
                rc = 1;
            }
            break;
        case 'h':
            demoUsage();
            rc = 1;
            break;
        default:
            printf( "Unrecognised flag:%s\n", argList[i] );
            demoUsage();
            rc = 1;
            break;
        }

        if ( rc == 1 ) {
            break;
        }
    }

    if ( rc == 2 ) {
        /* Mandatory parm missing */
        demoUsage();
    }
    
    return rc;
}

static int demoUsage( void ) {
    printf( "democ -a address <-s port> <-m mean rate> <-x max rate>\n" );
    printf( "     <-l min rate> <-r interval> <-p publish interval>\n" );
    printf( "     <-h>\n" );
    printf( "\n" );
    printf( "  -m: Mean rate of flow.\n" );
    printf( "  -x: Max rate of flow.\n" );
    printf( "  -l: Min rate of flow.\n" );
    printf( "  -r: Interval between flow rate changes (seconds).\n" );
    printf( "  -p: Interval between flow rate publication (seconds).\n" );
    printf( "  -a: TCP/IP address of MQIsdp broker e.g. 255.255.255.255\n" );
    printf( "  -s: TCP/IP port number of MQIsdp broker. Default 1883.\n" );
    printf( "  -h: Usage\n" );

    return 0;
}

static int demoMQIsdpConnect( PROG_PARMS *ppp ) {
    CONN_PARMS  cp;
    int         rc = MQISDP_OK;
    long        connState = MQISDP_CONNECTING;
    char        rcBuf[MQISDP_RC_STRING_LENGTH];

    /* ####### CONNECT ####### */
    /* Sizeof connect structure:            */
    /* sizeof CONN_PARMS                    */
    /* No additional MQIsdp brokers         */
    /* No Will message                      */

    cp.version = 1;
    cp.strucLength = sizeof(CONN_PARMS);
    
    /* Add a client identifier */
    strncpy( cp.clientId, CLIENT_ID, MQISDP_CLIENT_ID_LENGTH );

    cp.retryCount = 3;
    cp.retryInterval = 10;
    cp.options = MQISDP_NONE; /* | MQISDP_CLEAN_START; */
    cp.keepAliveTime = 15;
    cp.numServers = 1;
    
    /* Mailbox handles */
    cp.apiMailbox = ppp->MqisdpParms.hMbs.apiMailbox;
    cp.sendMailbox = ppp->MqisdpParms.hMbs.sendMailbox;
    cp.sendMutex = ppp->MqisdpParms.hMbs.sendMutex;
    cp.receiveSemaphore = ppp->MqisdpParms.hMbs.receiveSemaphore;
    
    /* Port number */
    cp.mqisdpSvr[0].portNumber = ppp->IpPort;
    
    /* IP address */
    strncpy( cp.mqisdpSvr[0].ipAddress, ppp->IpAddress, MQISDP_INET_ADDR_LENGTH );

    rc = MQIsdp_connect( &(ppp->MqisdpParms.hConn), &cp );
    demoGetRCtoString( rc, rcBuf, MQISDP_RC_STRING_LENGTH );
    demoLog( "MQIsdp_connect rc:%s\n", rcBuf );
    
    while ( connState == MQISDP_CONNECTING ) {
        connState = MQIsdp_status(ppp->MqisdpParms.hConn, MQISDP_INFO_STRING_LENGTH,
                                  &(ppp->MqisdpParms.infoCode), ppp->MqisdpParms.infoString );
        demoGetRCtoString( connState, rcBuf, MQISDP_RC_STRING_LENGTH );
        demoLog( "MQIsdp_status state:%s - %s\n", rcBuf, ppp->MqisdpParms.infoString ); 
        demoSleep(1000);
    }

    /* If we haven't successfully connected to the MQIsdp broker then disconnect */
    if ( connState != MQISDP_CONNECTED ) {
        demoMQIsdpDisconnect( ppp );
        rc = 1;
    }

    return rc;
}

static int demoMQIsdpPublish( PROG_PARMS *ppp ) {
    MQISDPMH   hMsg;
    PUB_PARMS *pPp;
    int        rc = MQISDP_OK;
    char       rcBuf[MQISDP_RC_STRING_LENGTH];

    pPp = (PUB_PARMS*)malloc( sizeof(PUB_PARMS) );

    if ( pPp != NULL ) {
        pPp->version = 1;
        pPp->strucLength = sizeof(PUB_PARMS);

        pPp->options = MQISDP_QOS_0;
    
        /* Add the topic name */
        pPp->topicLength = strlen(ppp->ControlTopic);
        pPp->topic = ppp->ControlTopic;
    
        /* Add the data */
        pPp->dataLength = strlen(ppp->Pairs);
        pPp->data = ppp->Pairs;
        
        /* Publish */
        rc = MQIsdp_publish( ppp->MqisdpParms.hConn, &hMsg, pPp );
        
        demoGetRCtoString( rc, rcBuf, MQISDP_RC_STRING_LENGTH );
        demoLog( "MQIsdp_publish rc:%s\n", rcBuf );
        
        free( pPp );
    }
    
    return rc;
}

static int demoMQIsdpDisconnect( PROG_PARMS *ppp ) {

    MQIsdp_disconnect( &(ppp->MqisdpParms.hConn) );

    return 0;
}


static void demoLog( char *fmt, ... ) {
    char       buf[128];
    va_list    arg_ptr;

    /* Should we log this message? mspLogOptions is defined at the top of this file */
    va_start( arg_ptr, fmt );
    vsprintf ( buf, fmt, arg_ptr );
    va_end( arg_ptr );

    printf( buf );
}

static void demoSleep( int mSecs ) {
    #ifdef WIN32
      Sleep( mSecs );
    #endif
    
    #ifdef UNIX
      struct timeval tv;
      tv.tv_sec = (int)(mSecs / 1000);
      tv.tv_usec = (mSecs % 1000) * 1000;
      select( 0, NULL, NULL, NULL, &tv );
    #endif
}

static int demoGetRCtoString( long rc, char *pBuffer, long bufSz ) {
    
    char *ptr = NULL;

    switch ( rc ) {
       /* Return codes */
       CaseRet( MQISDP_OK )
       CaseRet( MQISDP_Q_FULL )
       CaseRet( MQISDP_FAILED )
       CaseRet( MQISDP_PUBS_AVAILABLE )
       CaseRet( MQISDP_NO_PUBS_AVAILABLE )
       CaseRet( MQISDP_CONN_HANDLE_ERROR )
       CaseRet( MQISDP_MSG_HANDLE_ERROR )
       CaseRet( MQISDP_NO_WILL_TOPIC )
       CaseRet( MQISDP_INVALID_STRUC_LENGTH )
       CaseRet( MQISDP_DATA_LENGTH_ERROR )
       CaseRet( MQISDP_DATA_TOO_BIG )
       CaseRet( MQISDP_ALREADY_CONNECTED )
       CaseRet( MQISDP_CONNECTION_BROKEN )
       CaseRet( MQISDP_DATA_TRUNCATED )
       CaseRet( MQISDP_OUT_OF_MEMORY )
        
       /* Message states */
       CaseRet( MQISDP_DELIVERED )
       CaseRet( MQISDP_RETRYING )
       CaseRet( MQISDP_IN_PROGRESS )
       CaseRet( MQISDP_DISCARDED )
       CaseRet( MQISDP_CONNECTING )
       CaseRet( MQISDP_CONNECTED )
       CaseRet( MQISDP_DISCONNECTED )
        
       /* Info codes */
       CaseRet( MQISDP_PROTOCOL_VERSION_ERROR )
       CaseRet( MQISDP_CLIENT_ID_ERROR )
       CaseRet( MQISDP_BROKER_UNAVAILABLE )
       CaseRet( MQISDP_SOCKET_CLOSED )
        default:
            break;
    }
    
    if ( pBuffer != NULL ) {
        if ( ptr != NULL ) {
            strncpy( pBuffer, ptr, bufSz );
        } else {
            sprintf( pBuffer, "%ld", rc );
        }
    }
    
    return rc;
}

